#include "../calypso_api_i.h"

#ifdef CALYPSO_NO_TIME_T

#ifdef UNDER_CE
BOOL get_datetime(DATETIME_ST *dst)
{
  if (dst != NULL)
  {
	  SYSTEMTIME sys_time;
	  GetSystemTime(&sys_time);

    dst->year = sys_time.wYear;
    dst->month = (BYTE) sys_time.wMonth;
    dst->day = (BYTE) sys_time.wDay;
    dst->hour = (BYTE) sys_time.wHour;
    dst->minute = (BYTE) sys_time.wMinute;
    dst->second = (BYTE) sys_time.wSecond;
  }
  return TRUE;
}
#endif

BOOL is_date_valid(DATETIME_ST *dt)
{  
  if (dt == NULL) return FALSE;
      
  if (dt->year < 100)
  {
    if (dt->year > 70)
      dt->year += 1900;
    else
      dt->year += 2000;
  }
  
  if ((dt->month == 0) || (dt->month > 12)) return FALSE;
  if ((dt->day == 0)   || (dt->day   > 31)) return FALSE;  
  return TRUE;    
}

static BOOL _leap_year(WORD year)
{
  return (year % 4) == 0 && ((year % 100) != 0 || (year % 400) == 0);
}

void datetime_next_day(DATETIME_ST *dt)
{
  dt->day++;
    
  switch (dt->month)
  {     
    case 2 :
      /* February */
      if (_leap_year(dt->year))
      {
        if (dt->day <= 29) return;
      } else
      {
        if (dt->day <= 28) return;
      }
      break;
      
    /* Monthes with 30 days */
    case 4 : case 6 : case 9 : case 11 :
      if (dt->day <= 30) return;
      break;

    /* Monthes with 31 days */
    default:
      if (dt->day <= 31) return;
      break;
  }
    
  dt->day = 1;
  dt->month++;
  if (dt->month > 12)
  {
    dt->month = 1;
    dt->year++;
  }
}

void datetime_prev_day(DATETIME_ST *dt)
{
  if (dt->day <= 1)
  {
    if (dt->month <= 1)
    {
      /* Previous year */
      dt->year--;
      dt->month = 12;
    } else
    {
      /* Previous month */
      dt->month--;
    }

    switch (dt->month)
    {     
      case 2 :
        /* February */
        if (_leap_year(dt->year))
        {
          dt->day = 29;
        } else
        {
          dt->day = 28;
        }
        break;
        
      /* Monthes with 30 days */
      case 4 : case 6 : case 9 : case 11 :
        dt->day = 30;
        break;
  
      /* Monthes with 31 days */
      default:
        dt->day = 31;
        break;
    }    
  } else
  {
    /* Previous day */
    dt->day--;
  }
}

BOOL CalypsoDate14ToTimestamp(DATETIME_ST *dst, WORD src)
{
  if (dst == NULL) return FALSE;
  
  dst->hour   = 0;
  dst->minute = 0;
  dst->second = 0;
  
  /* 01/01/1997 */
  dst->day    = 1;
  dst->month  = 1;
  dst->year   = 1997;
  
  /* increment */
  while (src--)
    datetime_next_day(dst);    
  
  return is_date_valid(dst);
}

WORD CalypsoTimestampToDate14(const DATETIME_ST *src)
{
  DATETIME_ST timestamp;
  register WORD value = 0;
  
  if (src == NULL) goto done;

  memcpy(&timestamp, src, sizeof(DATETIME_ST));
  timestamp.hour   = 0;
  timestamp.minute = 0;
  timestamp.second = 0;

  if (!is_date_valid(&timestamp)) goto done;
  
  if (timestamp.year < 1997) goto done;
  
  while ((timestamp.year  != 1997)
      || (timestamp.month != 1)
      || (timestamp.day   != 1))
  {
    datetime_prev_day(&timestamp);
    value++;
  }
  
done:
  return value;  
}

BOOL CalypsoTime11ToTimestamp(DATETIME_ST *dst, WORD src)
{
  if (dst == NULL) return FALSE;

  dst->hour   = (BYTE) (src / 60);
  dst->minute = (BYTE) (src % 60);
  dst->second = 0;
  
  return TRUE;
}

WORD CalypsoTimestampToTime11(const DATETIME_ST *src)
{
  WORD value = 0;
  
  if (src == NULL) goto done;

  value  = src->hour;
  value *= 60;
  value += src->minute;

done:
  return value;
}

BOOL CalypsoDate14IsBefore(const DATETIME_ST *now, WORD value)
{
  DATETIME_ST chk, _now;

  if (now == NULL)
  {
    get_datetime(&_now);
    now = &_now;
  }
  
  CalypsoDate14ToTimestamp(&chk, value);
 
  if (chk.year  > now->year) return TRUE;
  if ((chk.year == now->year) && (chk.month > now->month)) return TRUE;
  if ((chk.year == now->year) && (chk.month == now->month) && (chk.day >= now->day)) return TRUE;

  return FALSE;
}

BOOL CalypsoDate14IsAfter(const DATETIME_ST *now, WORD value)
{
  DATETIME_ST chk, _now;

  if (now == NULL)
  {
    get_datetime(&_now);
    now = &_now;
  }

  CalypsoDate14ToTimestamp(&chk, value);
 
  if (chk.year  < now->year) return TRUE;
  if ((chk.year == now->year) && (chk.month < now->month)) return TRUE;
  if ((chk.year == now->year) && (chk.month == now->month) && (chk.day <= now->day)) return TRUE;

  return FALSE;
}

#endif
