#include <windows.h>
#include <winscard.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "pcsc.h"
#include "cardware/calypso/calypso_api.h"

void get_line(char *s, DWORD n);
P_CALYPSO_CTX ctx = NULL;

void exit_handler(void)
{
  PcscCleanup();

  if (ctx != NULL)
  {
    CalypsoDestroyContext(ctx);
    ctx = NULL;
  }

}

int main(int argc, char **argv)
{
  BYTE atr[32];
  DWORD atrlen;
  DWORD i, j, c, rc;
  DWORD card_idx = 0, sam_idx = 0;
  DWORD data_size;
  BYTE  data[256];
  BOOL ratified = FALSE;
  SCARDHANDLE card_handle;
  SCARDHANDLE sam_handle;
  char scan_buffer[32];

  atexit(exit_handler);

  for (i=1; i<argc; i++)
  {
    if (!strcmp("-v", argv[i]))
    {
      CalypsoSetTraceLevel(255);
      CalypsoSetTraceFile("stdout");
    }
  }

  ctx = CalypsoCreateContext();

  /* Start the PC/SC stuff */
  if (!PcscStartup())
  {
    printf("Failed to create PC/SC context");
    exit(EXIT_FAILURE);
  }

  /* Get the list of readers */
  c = PcscReaderCount();
  switch (c)
  {
    case 0  : printf("No PC/SC reader, exiting\n");
              exit(EXIT_SUCCESS);
              break;
    case 1  : printf("Found 1 PC/SC reader\n");
              break;
    default : printf("Found %ld PC/SC readers\n", c);
              break;
  }

  /* Display every reader */
  for (i=0; i<c; i++)
  {
    printf("[%ld] : %s\n", i+1, PcscReaderName(i));

    if (PcscCardPresent(i))
    {
      if (i == 0)
      {
        /* There's a card in the first reader, select it by default */
        card_idx = i+1;
      } else
      if (sam_idx == 0)
      {
        /* Select the first SAM */
        sam_idx = i+1;
      }

      /* Display the ATR (if some) */
      if (PcscCardAtr(i, atr, &atrlen))
      {
        printf("\tCard present, ATR=");
        for (j=0; j<atrlen; j++)
          printf("%02X", atr[j]);
        printf("\n");
      } else
      {
        printf("\tCard present, but no ATR available\n");
      }
    } else
      printf("\tNo card in the reader\n");
  }

  /* Ask the user to select the reader where the SAM is */
  scan_buffer[0] = '\0';
  for (;;)
  {
    if ((sam_idx > 0) && (sam_idx <= c)) sprintf(scan_buffer, "%ld", sam_idx);
    printf("\nSelect the PC/SC reader holding the CALYPSO SAM [%s] ", scan_buffer);
    get_line(scan_buffer, sizeof(scan_buffer));

    if (strlen(scan_buffer))
    {
      if ((scan_buffer[0] == 'q') || (scan_buffer[0] == 'Q')) exit(EXIT_FAILURE);
      sam_idx = atoi(scan_buffer);

      if ((sam_idx > 0) && (sam_idx <= c))
        if (PcscCardPresent(sam_idx - 1))
          break;

      printf("No smartcard in this reader - try again (Q to quit)\n");
    } else
    {
      if ((sam_idx > 0) && (sam_idx <= c))
        if (PcscCardPresent(sam_idx - 1))
          break;
    }
  } 

  sam_idx--;

  /* Connect to the SAM */
  if (!PcscConnect(sam_idx, &sam_handle))
  {
    printf("Failed to connect to the SAM.\n");
    exit(EXIT_FAILURE);
  }

  /* Try to bind our CALYPSO library to the card */
  rc = CalypsoSamBindPcsc(ctx, sam_handle);
  if (rc)
  {
    printf("CalypsoSamBindPcsc failed, error %04X.\n", rc);
    exit(EXIT_FAILURE);
  }

  printf("Connected to what may be a CALYPSO SAM ; we'll see later if this is right !\n");

  /* Ask the user to select the reader where the CALYPSO card is, or will be */
  scan_buffer[0] = '\0';
  for (;;)
  {
    if ((card_idx > 0) && (card_idx <= c)) sprintf(scan_buffer, "%ld", card_idx);
    printf("\nSelect the PC/SC reader holding the CALYPSO card [%s] ", scan_buffer);
    get_line(scan_buffer, sizeof(scan_buffer));
    
    if (strlen(scan_buffer))
      card_idx = atoi(scan_buffer);

    if ((card_idx > 0) && (card_idx <= c)) break;
  } 

  card_idx--;
  
  /* Confirm there's a card in the reader, if not we'll have to wait */
  if (!PcscCardPresent(card_idx))
  {
    /* Wait for a card */
    printf("Put your CALYPSO card on reader %s\n", PcscReaderName(card_idx));
    printf("(Ctrl+C to cancel)\n");

    if (!PcscWaitCard(card_idx, atr, &atrlen))
    {
      printf("Failed !\n");
      exit(EXIT_FAILURE);
    }

    printf("\tCard inserted, ATR=");
    for (j=0; j<atrlen; j++)
      printf("%02X", atr[j]);
    printf("\n");
  }

  /* Connect to the card */
  if (!PcscConnect(card_idx, &card_handle))
  {
    printf("Failed to connect to the card.\n");
    exit(EXIT_FAILURE);
  }

  /* Try to bind our CALYPSO library to the card */
  rc = CalypsoCardBindPcsc(ctx, card_handle);
  if (rc)
  {
    printf("CalypsoCardBindPcsc failed, error %04X.\n", rc);
    exit(EXIT_FAILURE);
  }

  /* Check this card is actually CALYPSO (ATR parsing or SELECT APPLICATION + FCI parsing) */
  rc = CalypsoCardActivate(ctx, NULL, 0);
  if (rc)
  {
    if (rc == CALYPSO_CARD_NOT_SUPPORTED)
    {
      printf("This is not a CALYPSO card !\n");
    } else
    {
      printf("CalypsoCardActivate failed, error %04X.\n", rc);
    }
    exit(EXIT_FAILURE);
  }

  /* OK, let's perform a ReadRecord transaction */
  /* ------------------------------------------ */
  printf("\nTrying READ RECORD transaction...\n");

  /* Initialise the timer */
  CalypsoBench(TRUE);

  rc = CalypsoStartTransactionEx(ctx, &ratified, CALYPSO_KEY_DEBIT, CALYPSO_KIF_DEBIT, 0, 0, NULL, NULL);
  if (rc)
  {
    printf("CalypsoStartTransactionEx failed, error %04X.\n", rc);
    exit(EXIT_FAILURE);
  }

  rc = CalypsoCardReadRecord(ctx, CALYPSO_SFI_ENVIRONMENT, 1, 19, data, &data_size);
  if (rc)
  {
    printf("CalypsoCardReadRecord in transaction failed, error %04X.\n", rc);
    exit(EXIT_FAILURE);
  }

  rc = CalypsoCommitTransaction(ctx, TRUE);
  if (rc)
  {
    printf("CalypsoCommitTransaction in transaction failed, error %04X.\n", rc);
    exit(EXIT_FAILURE);
  }

  /* Done ! */
  printf("READ RECORD transaction performed in %ldms\n", CalypsoBench(TRUE));
  printf("Last transaction was %sratified\n", ratified ? "" : "NOT ");



  /* Drop card and SAM connections */
  CalypsoCardDispose(ctx);
  PcscDisconnect(card_handle, SCARD_EJECT_CARD);
  CalypsoSamDispose(ctx);
  PcscDisconnect(sam_handle, SCARD_LEAVE_CARD);

  exit(EXIT_SUCCESS);
}




void get_line(char *s, DWORD n)
{
  DWORD w = 0;
  int c;

  memset(s, '\0', n);
  while ((c = fgetc (stdin)) != '\n' && c != EOF)
  {
    if (w < (n - 1))
    {
      s[w] = c;
      w++;
    }
  }
}
