/**h* NFCTools/nfc_rtd_text.c
 *
 * NAME
 *   nfc_rtc_text.c
 *
 * DESCRIPTION
 *   Implementation of the NFC RTD Text
 *
 * COPYRIGHT
 *   (c) 2011 PRO ACTIVE SAS - See LICENCE.txt for licence information
 *
 * AUTHOR
 *   Johann Dantant / PRO ACTIVE for www.springcard.com
 *
 * HISTORY
 *   JDA 17/01/2011 : first public release
 *
 **/
#include "nfc.h"

BOOL text_decode_alloc_a(char **dst_text, char **dst_lang, unsigned char *src, size_t src_sz)
{
  unsigned char header;
  size_t lang_sz, text_sz;

  if ((src == NULL) || (src_sz < 1))
    return FALSE;

  if (dst_text != NULL)
    *dst_text = NULL;
  if (dst_lang != NULL)
    *dst_lang = NULL;

  header = src[0];

  if (header & 0x80)
  {
    /* UTF-16 string */
    if (!nfc_silent)
      printf("Sorry, UTF-16 is not supported yet\n");
    return FALSE;
  } else
  {
    /* UTF-8 string */
  }

  if (header & 0x40)
  {
    /* RFU ! */
  }

  /* Compute sizes */
  lang_sz = header & 0x3F;
  if (lang_sz > (src_sz - 1))
  {
    if (!nfc_silent)
      printf("Invalid Text RTD\n");
    return FALSE; /* Oups */
  }
  text_sz = src_sz - lang_sz - 1;

  /* Copy the lang */
  if ((lang_sz != 0) && (dst_lang != NULL))
    *dst_lang = nfc_strdup_blob(&src[1], lang_sz);

  /* Copy the actual text */
  if ((text_sz != 0) && (dst_text != NULL))
    *dst_text = nfc_strdup_blob(&src[1 + lang_sz], text_sz);

  return TRUE;
}

BOOL text_encode_alloc_a(unsigned char **dst, size_t *dst_sz, const char *text, const char *lang)
{
  unsigned char header;
  size_t lang_sz = 0;
	size_t text_sz = 0;

  header = 0x00;

  if (lang != NULL)
	{
	  lang_sz  = strlen(lang);
		lang_sz &= 0x3F;
		header  |= lang_sz;
	}

	if (text != NULL)
    text_sz = strlen(text);

  if (dst_sz != NULL)
	  *dst_sz = 1 + lang_sz + text_sz;

	if (dst != NULL)
	{
	  unsigned char *_dst = nfc_alloc(1 + lang_sz + text_sz);

		_dst[0] = header;

		if (lang_sz)
		  memcpy(&_dst[1], lang, lang_sz);
		if (text_sz)
		  memcpy(&_dst[1 + lang_sz], text, text_sz);

		*dst = _dst;
	}

  return TRUE;
}

