/*
  THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
  ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED
  TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
  PARTICULAR PURPOSE.

  Copyright (c) 2003-2011 PRO ACTIVE SAS - www.springcard.com
*/

#include "springprox_pcsc.h"

#ifdef WIN32
#pragma comment(lib,"winscard.lib")
#endif

BOOL GetReaderList(char ***aszReaders)
{
	SCARDCONTEXT      hContext;
	LPSTR             szReaders = NULL;
  DWORD             dwReadersSz;
	LPTSTR            pReader;
	int               cReaders, i;
  LONG              rc;
	char              **lReaders;

	if (aszReaders == NULL)
	  return FALSE;

	rc = SCardEstablishContext(SCARD_SCOPE_SYSTEM, NULL, NULL, &hContext);
	if(rc != SCARD_S_SUCCESS)
	{
		printf("SCardEstablishContext error %08lX\n", rc);
		return FALSE;
	}

  dwReadersSz = SCARD_AUTOALLOCATE;
	rc = SCardListReaders(hContext,
                        NULL,                /* Any group */
                        (LPTSTR) &szReaders, /* Diabolic cast for buffer auto-allocation */
                        &dwReadersSz);       /* Beg for auto-allocation */
	  
	if (rc == SCARD_E_NO_READERS_AVAILABLE)
	{
		/* No reader at all */
		SCardReleaseContext(hContext);
		return TRUE;
	}

  if (rc != SCARD_S_SUCCESS)
	{
    printf("SCardListReaders error %08lX\n",rc);
		SCardReleaseContext(hContext);
		return FALSE;
	}

	if (szReaders == NULL)
	{
    printf("SCardListReaders returned a NULL pointer\n");
		SCardReleaseContext(hContext);
		return FALSE;
	}

  /* Count the readers */
	cReaders = 0;
  pReader  = szReaders;
	while (*pReader != '\0')
	{
	  cReaders++;
    /* Jump to next entry in multi-string array */
		pReader += strlen(pReader) + 1;
	}

	if (!cReaders)
	{
		/* No reader at all */
    SCardFreeMemory(hContext, szReaders);
		SCardReleaseContext(hContext);
		return TRUE;
	}

  /* Allocate a list of pointers (cReader+1 to terminate the list by NULL) */
	lReaders = calloc(cReaders+1, sizeof(void *));
	if (lReaders == NULL)
	{
	  printf("Memory allocation failed\n");
		SCardFreeMemory(hContext, szReaders);
		SCardReleaseContext(hContext);
		return FALSE;
	}

  /* Populate the list */
  i = 0;
  pReader = szReaders;
	while (*pReader != '\0')
	{
	  /* Allocate and copy the name */
	  lReaders[i] = strdup(pReader);

		if (lReaders[i]  == NULL)
		{
			printf("Memory allocation failed\n");
			FreeReaderList(lReaders);
			SCardFreeMemory(hContext, szReaders);
			SCardReleaseContext(hContext);
			return FALSE;
		}

    /* Jump to next entry in multi-string array */
		pReader += strlen(pReader) + 1;
	  i++;
	}

  /* Done */
	SCardFreeMemory(hContext, szReaders);
	SCardReleaseContext(hContext);
	*aszReaders = lReaders;
	return TRUE;
}

void FreeReaderList(char **aszReaders)
{
  int i;

  if (aszReaders != NULL)
	{
	  for (i=0; aszReaders[i] != NULL; i++)
		{
			free(aszReaders[i]);
		}

		free(aszReaders);
	}
}

BOOL PrintReaderStatus(const char *szReaderName)
{
	SCARDCONTEXT      hContext;
  SCARD_READERSTATE rgscState;
	unsigned int i;
	LONG rc;

	if (szReaderName == NULL)
	  return FALSE;

	rc = SCardEstablishContext(SCARD_SCOPE_SYSTEM, NULL, NULL, &hContext);
	if(rc != SCARD_S_SUCCESS)
	{
		printf("SCardEstablishContext error %08lX\n", rc);
		return FALSE;
	}

  /* Get status */
  rgscState.szReader = szReaderName;
  rgscState.dwCurrentState = SCARD_STATE_UNAWARE;

 	rc = SCardGetStatusChange(hContext,
			                      0,
			                      &rgscState,
			                      1);	

  if (rc != SCARD_S_SUCCESS)
	{
    printf("SCardGetStatusChange(%s) error %08lX\n", szReaderName, rc);
		SCardReleaseContext(hContext);
		return FALSE;
	}

  /* Show reader's status */
  if (rgscState.dwEventState & SCARD_STATE_IGNORE)
    printf("\tIgnore this reader\n");

  if (rgscState.dwEventState & SCARD_STATE_UNKNOWN)
    printf("\tReader unknown\n");

  if (rgscState.dwEventState & SCARD_STATE_UNAVAILABLE)
    printf("\tStatus unavailable\n");

  if (rgscState.dwEventState & SCARD_STATE_EMPTY)
    printf("\tNo card in the reader\n");

  if (rgscState.dwEventState & SCARD_STATE_PRESENT)
    printf("\tCard present\n");

  /* Show ATR (if some) */
  if (rgscState.cbAtr)
	{
		printf("\tCard ATR: ");
		for (i=0; i<rgscState.cbAtr; i++)
			printf("%02X", rgscState.rgbAtr[i]);
    printf("\n");
  }

  /* Show card's status */
  if (rgscState.dwEventState & SCARD_STATE_ATRMATCH)
    printf("\tATR match\n");

  if (rgscState.dwEventState & SCARD_STATE_INUSE)
    printf("\tCard (or reader) in use\n");

  if (rgscState.dwEventState & SCARD_STATE_EXCLUSIVE)
    printf("\tCard (or reader) reserved for exclusive use\n");

  if (rgscState.dwEventState & SCARD_STATE_MUTE)
    printf("\tCard is mute\n");

  SCardReleaseContext(hContext);
  return TRUE;
}

BOOL ReaderGetStatus(const char *szReaderName, DWORD *pdwStatus)
{
	SCARDCONTEXT      hContext;
  SCARD_READERSTATE rgscState;
	LONG rc;

	if (szReaderName == NULL)
	  return FALSE;

	rc = SCardEstablishContext(SCARD_SCOPE_SYSTEM, NULL, NULL, &hContext);
	if(rc != SCARD_S_SUCCESS)
	{
		printf("SCardEstablishContext error %08lX\n", rc);
		return FALSE;
	}

  /* Get status */
  rgscState.szReader = szReaderName;
  rgscState.dwCurrentState = SCARD_STATE_UNAWARE;

 	rc = SCardGetStatusChange(hContext,
			                      0,
			                      &rgscState,
			                      1);	

  if (rc != SCARD_S_SUCCESS)
	{
    printf("SCardGetStatusChange(%s) error %08lX\n", szReaderName, rc);
		SCardReleaseContext(hContext);
		return FALSE;
	}

  *pdwStatus = rgscState.dwEventState;

  SCardReleaseContext(hContext);
  return TRUE;
}

BOOL CardAvailable(const char *szReaderName)
{
  DWORD dwStatus;

	if (!ReaderGetStatus(szReaderName, &dwStatus))
	  return FALSE;

  return ((dwStatus & SCARD_STATE_PRESENT) && !(dwStatus & SCARD_STATE_MUTE) && !(dwStatus & SCARD_STATE_INUSE));
}

BOOL CardConnect(const char *szReaderName, CARD_CHANNEL_T *pChannel)
{
  LONG rc;

  if (pChannel == NULL)
	  return FALSE;

	rc = SCardEstablishContext(SCARD_SCOPE_SYSTEM, NULL, NULL, &pChannel->hContext);
	if(rc != SCARD_S_SUCCESS)
	{
		printf("SCardEstablishContext error %08lX\n", rc);
		return FALSE;
	}

  /* Connect to the card */
  rc = SCardConnect(pChannel->hContext, szReaderName, SCARD_SHARE_EXCLUSIVE, SCARD_PROTOCOL_T0|SCARD_PROTOCOL_T1, &pChannel->hCard, &pChannel->dwProtocol);
  if (rc != SCARD_S_SUCCESS)
	{
    printf("SCardConnect(%s) error %08lX\n", szReaderName, rc);
		SCardReleaseContext(pChannel->hContext);
		return FALSE;
	}

  pChannel->dwAtrLength = sizeof(pChannel->cbAtr);
	rc = SCardStatus(pChannel->hCard, NULL, NULL, NULL, NULL, pChannel->cbAtr, &pChannel->dwAtrLength);
  if (rc != SCARD_S_SUCCESS)
	{
    printf("SCardStatus error %08lX\n", rc);
		SCardDisconnect(pChannel->hCard, SCARD_EJECT_CARD);
		SCardReleaseContext(pChannel->hContext);
		return FALSE;
	}

  return TRUE;
}

BOOL CardTransmit(CARD_CHANNEL_T *pChannel, const BYTE *abCommandApdu, DWORD dwCommandApduLength, BYTE *abResponseApdu, DWORD *dwResponseApduLength)
{
  LONG rc;
	const SCARD_IO_REQUEST *pioSendPci;

  if (pChannel == NULL)
	  return FALSE;

	switch (pChannel->dwProtocol)
	{
	  case SCARD_PROTOCOL_T0 : pioSendPci = SCARD_PCI_T0; break;
    case SCARD_PROTOCOL_T1 : pioSendPci = SCARD_PCI_T1; break;
		default                : pioSendPci = NULL;
	}

  rc = SCardTransmit(pChannel->hCard, pioSendPci, abCommandApdu, dwCommandApduLength, NULL, abResponseApdu, dwResponseApduLength);
  if (rc != SCARD_S_SUCCESS)
	{
    printf("SCardTransmit error %08lX\n", rc);
		return FALSE;
	}

  return TRUE;
}

BOOL CardDisconnect(CARD_CHANNEL_T *pChannel, DWORD dwDisposition)
{
  LONG rc;

  if (pChannel == NULL)
	  return FALSE;

	rc = SCardDisconnect(pChannel->hCard, dwDisposition);
  if (rc != SCARD_S_SUCCESS)
	{
    printf("SCardDisconnect error %08lX\n", rc);
		SCardReleaseContext(pChannel->hContext);
		return FALSE;
	}

	rc = SCardReleaseContext(pChannel->hContext);
  if (rc != SCARD_S_SUCCESS)
	{
    printf("SCardReleaseContext error %08lX\n", rc);
		return FALSE;
	}

  return TRUE;
}

