using System;
using System.Collections;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.IO;
using System.Xml;
using System.Windows.Forms;

using SpringCardPCSC;

namespace CalypsoExplorer
{
  public partial class fMain : Form
  {
    private SCardReader reader = null;
    private StringCollection CardData;
    private string SelectedReader = "";
    private uint SelectedReaderStatus = 0;
    private bool CardWasPresent = false;
    private bool CardInsertionMutex = false;

    delegate void ReaderEventCallback();
  
    public fMain()
    {

      Splash MySplashForm = new Splash();
      MySplashForm.Show();
      InitializeComponent();
    }
  
    private void Display(XmlTextReader xml)
    {
      int i;
      TreeNode rootNode = null;
      TreeNode currentNode = null;
      string currentPath = "";
         
      treeView.BeginUpdate();
      treeView.Nodes.Clear();
      
      CardData = new StringCollection();
           
      while (xml.Read())
      {
        switch (xml.NodeType)
        {
          case XmlNodeType.Element :
            {
              string text = xml.Name;
              
              currentPath += "/" + xml.Name;
              
              if (xml.HasAttributes)
              {
                for (i=0; i<xml.AttributeCount; i++)
                {
                  xml.MoveToAttribute(i);

                  string name = xml.Name;
                  string value = xml.Value;
                  
                  if (name == "id")
                  {
                    currentPath += " #" + value;
                    text += " #" + value;
                  }
                }
              }
                         
              if (currentNode == null)
              {
                currentNode = treeView.Nodes.Add(currentPath, text);
                rootNode = currentNode;
              } else
              {
                currentNode = currentNode.Nodes.Add(currentPath, text);              
              }          
              
              if ( xml.IsEmptyElement)
                currentNode = currentNode.Parent;
            }
            break;
          case XmlNodeType.EndElement :
            {
              currentNode = currentNode.Parent;
              
              if (currentNode != null)
              {
                do
                {
                  for (i=0; i<currentNode.Nodes.Count; i++)
                  {
                    if (currentNode.Nodes[i].Text == "####LEAF####")
                    {
                      currentNode.Nodes.RemoveAt(i);
                      break;
                    }
                  }
                } while (i<currentNode.Nodes.Count);
              }
              
              currentPath = currentPath.Substring(0, currentPath.LastIndexOf("/"));
            }                        
            break;

          case XmlNodeType.Text :
          case XmlNodeType.CDATA :
            {
              currentNode.Text = "####LEAF####";
              CardData.Add(currentPath + "=" + xml.Value);
            }
            break;
        }
      }
    
      rootNode.Expand();
      treeView.EndUpdate();
    }

   
    private void OnSelectXmlNode(string path)
    {
      int i;
     
      listView.BeginUpdate();
      listView.Items.Clear();
      
   
      for (i=0; i<CardData.Count; i++)
      {
        string key = CardData[i].Substring(0, CardData[i].IndexOf('='));

        if (key.Length < path.Length) continue;  
        if (key.Substring(0, key.LastIndexOf('/')) != path) continue;

        key = key.Substring(key.LastIndexOf('/') + 1);
        
        string value = CardData[i].Substring(CardData[i].IndexOf('=')+1);

        ListViewItem listItem = listView.Items.Add(key);

        listItem.SubItems.Add(value);
      }
      
      listView.EndUpdate();
    }

    private void listView_Resize(object sender, EventArgs e)
    {
      int w = listView.Width - listView.Columns[0].Width - 5;
      if (w < 284) w = 284;
      listView.Columns[1].Width = w;
    }

    private void fMain_Shown(object sender, EventArgs e)
    {
      DisplayStatus();
      SelectReader();
    }
    
    private void SelectReader()
    {
      if (reader != null)
      {
          reader.StopMonitor();
          reader = null;
      }
        
      fReaderSelect ReaderSelectForm = new fReaderSelect();
      ReaderSelectForm.ShowDialog();      
      switch (ReaderSelectForm.DialogResult)
      {
        case DialogResult.OK :
          {
            SelectedReader = ReaderSelectForm.SelectedReader();
          }
          break;
        case DialogResult.Cancel :
        default :
          {          
          }
          break;      
      }
           
      if (SelectedReader == "")
      {
        toolStripStatusReader.Text = "Offline mode";
        toolStripStatusCard.Text = "";
      } else
      {
        toolStripStatusReader.Text = SelectedReader;
        toolStripStatusCard.Text = "Reading reader's status...";

        reader = new SCardReader(SelectedReader);
        reader.StartMonitor(OnReaderStatusChanged);
      }
    }
    
    private void OnReaderStatusChanged(uint State, CardBuffer Atr)
    {
        SelectedReaderStatus = State;
        this.Invoke(new ReaderEventCallback(DisplayStatus), null);

        if ((State & SCARD.STATE_UNAVAILABLE) != 0)
        {
          this.Invoke(new ReaderEventCallback(ReaderFatalError), null);
        } else
        {
          this.Invoke(new ReaderEventCallback(DisplayStatus), null);
          
          if ((State & SCARD.STATE_PRESENT) != 0)
          {
            if (!CardWasPresent && !CardInsertionMutex)
            {
              CardWasPresent = true;
              CardInsertionMutex = true;
              this.Invoke(new ReaderEventCallback(ReadCardAndDisplayData), null);
            }
          } else
          {
            CardWasPresent = false;
          }
        }
    }
    
    private void fMain_FormClosed(object sender, FormClosedEventArgs e)
    {
        if (reader != null)
        {
            reader.StopMonitor();
            reader = null;
        }
    }

    private void DisplayStatus()
    {
        toolStripStatusCard.Text = SCARD.ReaderStatusToString(SelectedReaderStatus);
          
      imgReaderInUse.Hide();
      imgReaderPresent.Hide();
      imgReaderError.Hide();
      imgReaderEmpty.Hide();            

      if (SelectedReader == "")
      {

      } else    
      if (toolStripStatusCard.Text.Contains("INUSE"))
      {
        imgReaderInUse.Show();
      } else      
      if (toolStripStatusCard.Text.Contains("PRESENT"))
      {
        imgReaderPresent.Show();
      } else
      if (toolStripStatusCard.Text.Contains("EMPTY"))
      {
        imgReaderEmpty.Show();
      } else
      {
        imgReaderError.Show();
      }      
    }

    private void ReaderFatalError()
    {
      MessageBox.Show("The reader has been removed from the system, or a fatal error has occured. Please try to connect a different reader.", "Reader error");      
    }
    
    private void ReadCardAndDisplayData()
    {
      SCardChannel card = new SCardChannel(SelectedReader);
      
      if (!card.Connect())
      {
        MessageBox.Show("Connection to the card has failed. Please make sure that the card has not been removed from the reader, and that no other program tries to access the card at the same time.", "Failed to read the card");
        CardInsertionMutex = false;
        return;
      }


      SCardCalypso calypso = new SCardCalypso(card);

      if (!calypso.Activated())
      {
        card.Disconnect();
        MessageBox.Show("The Calypso application has not been found in the card. Please make sure the card you put on the reader has got a valid '1TIC.ICA' application.", "Failed to read the card");
        CardInsertionMutex = false;
        return;      
      }
      
      string xml_data = calypso.ParseToXml();
     
      card.Disconnect();      

      if (xml_data == "")
      {
        MessageBox.Show("An error has occured while reading the data from the Calypso card. Please try again.", "Failed to read the card");
        CardInsertionMutex = false;
        return;
      }

      Display(new XmlTextReader(new StringReader(xml_data)));
      
      CardInsertionMutex = false;
    }

    private void exitToolStripMenuItem_Click(object sender, EventArgs e)
    {
      AboutBox aboutFrm = new AboutBox();
      aboutFrm.ShowDialog();
    }

    private void exitToolStripMenuItem1_Click(object sender, EventArgs e)
    {
      Close();
    }

    private void changeReaderToolStripMenuItem_Click(object sender, EventArgs e)
    {
      SelectReader();
    }

    private void readCardToolStripMenuItem_Click(object sender, EventArgs e)
    {
      ReadCardAndDisplayData();
    }

    private void treeView_AfterSelect(object sender, TreeViewEventArgs e)
    {
      OnSelectXmlNode(e.Node.Name);
    }    
    
  }  
}
