using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using SpringCardPCSC;
using Microsoft.Win32;
using System.Collections.ObjectModel;

namespace NFCTool
{
  public partial class ChooseReaderFrm : Form
  {
    // settings chosen by user
    public string ReaderName = "";
    private bool canClose = true;
    public bool connect = false;

    public PosterFrm parentWnd;

    public ChooseReaderFrm(PosterFrm CallingWindow, bool Closable)
    {
      InitializeComponent();
      canClose = Closable;
      buttonOk.DialogResult = DialogResult.OK;
      // reader Name ComboBox filled with list loaded with SPCSC.getListReaders

      string[] cReaders = SCARD.Readers;
      foreach (string cName in cReaders)
      {
        ReadercomboBox.Items.Add(cName);
      }

      if (cReaders.Length > 0)
      {
        ReadercomboBox.SelectedItem = 0;
      }

      ReadRegister();

      checkBoxAuto.Checked = connect;

      this.parentWnd = CallingWindow;
    }

    private void ReaderComboBox_SelectedIndexChanged(object sender, EventArgs e)
    {
      this.ReaderName = ReadercomboBox.SelectedItem.ToString();
    }


    public void ReadRegister()
    {
      // settings according to config registered
      try
      {
        RegistryKey confKey =
         Registry.CurrentUser.OpenSubKey("SOFTWARE\\SpringCard\\NFCTool");
        string[] ConfkeyNames = confKey.GetValueNames();
        foreach (string key in ConfkeyNames)
        {
          // setting former Reader name 
          if (key.Contains("Reader"))
          {
            if (ReadercomboBox.Items.Contains(confKey.GetValue(key)))
            {
              ReadercomboBox.SelectedItem = confKey.GetValue(key);
              ReaderName = confKey.GetValue(key).ToString();
            }

          }

          // setting former Reader name 
          if (key.Contains("AutoConnect"))
          {
            connect = true;
          }

        }
      }
      catch (Exception)
      {
        // do nothing, the key doesn't exist in register
      }
    }

    private void butOk_Click(object sender, EventArgs e)
    {
      this.parentWnd.SetReader(this.ReaderName);
      RegisterConfig();
      this.Close();
      this.Dispose();
    }



    // Saving current config into registry 
    private void RegisterConfig()
    {
        try
        {

            RegistryKey RecordKey =
            Registry.CurrentUser.CreateSubKey("SOFTWARE\\SpringCard\\NFCTool");

            // register reader values
            RecordKey.SetValue("Reader", ReaderName);

            if (checkBoxAuto.Checked)
            {
                RecordKey.SetValue("AutoConnect", "1");
            }
            else
            {
                RecordKey.DeleteValue("AutoConnect", false);
            }
        }
        catch (Exception)
        {
            // cannot write registry
        }

    }

    private void ChooseReaderFrm_FormClosing(object sender, FormClosingEventArgs e)
    {
      if (this.DialogResult != DialogResult.OK && canClose)
      {
        // close everything in case the form wasn't closed by "OK" button
        parentWnd.Close();
        parentWnd.Dispose();
      }
    }

    private void pictureBoxClose_Click(object sender, EventArgs e)
    {
      this.Close();
      this.Dispose();
    }

    private void pBReload_Click(object sender, EventArgs e)
    {
      // re load readers' list
      // erase former list
      ReadercomboBox.Items.Clear();
      string[] cReaders = SCARD.Readers;
      foreach (string cName in cReaders)
      {
        ReadercomboBox.Items.Add(cName);
      }
    }


  }
}