using System;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using SpringCardPCSC;


namespace NFC_Decoder
{
  public partial class NFCDecoderFrm : Form
  {
    SCardReader reader = null;
    SCardChannel scard = null;
    
    public NFCDecoderFrm()
    {
      InitializeComponent();
      set_URI_Codes();
    }

#region form management

    private void NFCDecoderFrm_Shown(object sender, EventArgs e)
    {
      /* Showing configuration form - User has to select reader */
      ConfigReaderForm cf = new ConfigReaderForm(this, true);
      cf.ReadRegister();

      if ((cf.ReaderName == "") || (!cf.connect))
        cf.ShowDialog();
      else
        SetReader(cf.ReaderName);
    }
    private void NFCDecoderFrm_FormClosed(object sender, FormClosedEventArgs e)
    {
      Cursor = Cursors.WaitCursor;

      if (scard != null)
      {
        scard.Disconnect();
        scard = null;
      }

      if (reader != null)
      {
        reader.StopMonitor();
        reader = null;
      }

      Cursor = Cursors.Default;
    }

    private void buttonYes_Click(object sender, EventArgs e)
    {
      panelPrompt.Visible = false;
      /* open URL */
      RunUrl(Card_URI_Tot, true);
      labelInfo.Text = "Opening URL " + Card_URI_Tot + " , please wait ...";
    }

    private void buttonNo_Click(object sender, EventArgs e)
    {
      panelPrompt.Visible = false;
    }

    bool alwaysOpen = false;
    bool neverOpen = false;

    private void checkBoxNeverOpen_CheckedChanged(object sender, EventArgs e)
    {
      if (checkBoxNeverOpen.Checked)
      {
        if (checkBoxAlwaysOpen.Checked)
          checkBoxAlwaysOpen.Checked = false;
        buttonNo.Visible = false;
        buttonYes.Visible = false;
        alwaysOpen = false;
        neverOpen = true;
      }
      else
      {
        if (checkBoxAlwaysOpen.Checked == false)
        {
          buttonYes.Visible = true;
          buttonNo.Visible = true;
        }
        neverOpen = false;
      }
    }

    private void checkBoxAlwaysOpen_CheckedChanged(object sender, EventArgs e)
    {
      if (checkBoxAlwaysOpen.Checked)
      {
        if (checkBoxNeverOpen.Checked)
          checkBoxNeverOpen.Checked = false;
        buttonNo.Visible = false;
        buttonYes.Visible = false;
        alwaysOpen = true;
        neverOpen = false;
      }
      else
      {
        if (checkBoxNeverOpen.Checked == false)
        {
          buttonYes.Visible = true;
          buttonNo.Visible = true;
        }
        alwaysOpen = false;
      }
    }

#endregion

#region reader management

    public void SetReader(string readerName)
    {
      reader = new SCardReader(readerName);
      reader.StartMonitor(new SCardReader.StatusChangeCallback(OnReaderStatusChanged));
    }

    bool processing_card = false;
    delegate void ReaderStatusChangedInvoker(uint ReaderState, CardBuffer CardAtr);

    int cpt_sector = 0;

    void OnReaderStatusChanged(uint ReaderState, CardBuffer CardAtr)
    {
      if (this.InvokeRequired)
      {
        this.BeginInvoke(new ReaderStatusChangedInvoker(OnReaderStatusChanged), ReaderState, CardAtr);
        return;
      }

      if (reader.CardPresent)
      {
        if (!processing_card)
        {

          processing_card = true;

          scard = new SCardChannel(reader.Name);
          labelInfo.Text = "Card present, start processing...";

          scard.ShareModeAsString = "SHARED";

          scard.Connect();
          if (!scard.Connected)
          {
            MessageBox.Show("Error " + scard.LastErrorAsString + " - try another card");
            return;
          }
          loadKeys();
          

        }
      }
      else
      {
        processing_card = false;
        setDefaults();
        labelInfo.Text = "Waiting for a card";
      }


    }

    void setDefaults()
    {
      cpt_sector = 0;
      processing_card = false;
      NFC_Content = false;
      blockToRead = 0;
      nb_sectors_nfc = 0;
      Card_URI_Tot = "";
      panelPrompt.Visible = false;
    }
#endregion

    public void processCard()
    {
      GetCardType();
      if(card_type != TYPE_UNSUPPORTED)
        ReadCard();
    }

#region load keys

    bool keys_loaded = false;
    NameValueCollection keys = new NameValueCollection();
    string[] keys_indexes = { "00", "01", "02", "10", "11", "12" };
    int cpt_index = 0;

    void loadKeys()
    {
      if (!keys_loaded)
      {
        keys.Set("00", "FF FF FF FF FF FF");
        keys.Set("10", "FF FF FF FF FF FF");
        keys.Set("01", "A0 A1 A2 A3 A4 A5");
        keys.Set("02", "D3 F7 D3 F7 D3 F7");
        keys.Set("11", "D3 F7 D3 F7 D3 F7");
        keys.Set("12", "53 43 4E 46 43 42");
        CAPDU capduLoadKey = new CAPDU("FF 82 00 00 06 FF FF FF FF FF FF");
        scard.Transmit(capduLoadKey, new SCardChannel.TransmitDoneCallback(CallbackLoadKey));

      }
      else
      {
        processCard();
      }
    }



    delegate void TransmitKeyDoneInvoker(RAPDU rapdu);
    void CallbackLoadKey(RAPDU rapdu)
    {
      if (this.InvokeRequired)
      {
        this.BeginInvoke(new TransmitKeyDoneInvoker(CallbackLoadKey), rapdu);
        return;
      }

      if (rapdu == null)
      {
        /* An error has occured */
        labelInfo.Text = "Error " + scard.LastErrorAsString;
      }
      else
      {
        if (rapdu.SW1 == 0x90 && rapdu.SW2 == 0x00)
        {
          cpt_index++;
          if (cpt_index < keys_indexes.Length)
          {
            CAPDU capduLoadKey = new CAPDU("FF 82 00 " + keys_indexes[cpt_index] + " 06 " + keys[keys_indexes[cpt_index]]);
            scard.Transmit(capduLoadKey, new SCardChannel.TransmitDoneCallback(CallbackLoadKey));
          }
          else
          {
            keys_loaded = true;
            processCard(); 
          }
        }
        else
        {
          labelInfo.Text = "Loading keys unsuccessful";
          processCard();
        }
      }

    }

#endregion

#region TLV decoding

    // TLV Constants

    const int INDEX_T = 0;
    const int INDEX_L = 1;
    const byte NULL_TLV = 0x00;
    const byte LOCK_CONTROL_TLV = 0x01;
    const byte MEMORY_CONTROL_TLV = 0x02;
    const byte NDEF_MESSAGE_TLV = 0x03;
    const byte PROPRIETARY_TLV = 0xFD;
    const byte TERMINATOR_TLV = 0xFE;

    const byte FLAG_L_3_BYTES = 0xFF;

    const int MAX_L = 65534;//FFFE

    string T_STR = "";
    int L_VALUE = 0;
    byte[] V_VALUE;


    private void decodeTLV(byte[] dataIn)
    {
      V_VALUE = new byte[0];
      if (dataIn.Length > 2)
      {
        // T
        byte t = dataIn[INDEX_T];
        T_STR = decodeT(t);

        // L
        int start_v = INDEX_L + 1;
        if (dataIn[INDEX_L] == FLAG_L_3_BYTES)
        {
          start_v += 2;
          L_VALUE = (int)(dataIn[INDEX_L + 1] << 8 + dataIn[INDEX_L + 2]);
        }
        else
        {
          L_VALUE = (int)dataIn[INDEX_L];
        }

        // V
        V_VALUE = new byte[L_VALUE];
        for (int cpt = 0; cpt < L_VALUE; cpt++)
        {
          if (dataIn.Length > cpt + start_v)
            V_VALUE[cpt] = dataIn[cpt + start_v];
        }
      }

    }


    private string decodeT(byte t)
    {
      switch (t)
      {
        case NULL_TLV: return "NULL";
        case LOCK_CONTROL_TLV: return "Lock Control";
        case MEMORY_CONTROL_TLV: return "Memory Control";
        case NDEF_MESSAGE_TLV: return "NDEF Message";
        case PROPRIETARY_TLV: return "Proprietary";
        case TERMINATOR_TLV: return "Terminator";
        default: return "Type unknown";
      }
    }


#endregion

#region NDEF decoding

    const byte MESSAGE_BEGIN_FLAG = 0x80;
    const byte MESSAGE_END_FLAG = 0x40;
    const byte CHUNK_FLAG = 0x20;
    const byte SHORT_RECORD_FLAG = 0x10;
    const byte ID_LENGTH_FLAG = 0x08;

    const byte SHORT_RECORD_HEADER = 0xD1;
    const byte SR_MB_HEADER = 0x91;
    const byte SR_ME_HEADER = 0x51;
    const byte RECORD_HEADER = 0xC1;

    // TNF constants
    const byte EMPTY = 0x00;
    const byte NFC_TYPE = 0x01;
    const byte MEDIA_TYPE = 0x02;
    const byte ABSOLUTE_URI = 0x03;
    const byte EXTERNAL_TYPE = 0x04;
    const byte UNKNOWN_TYPE = 0x05;
    const byte UNCHANGED = 0x06;
    const byte RESERVED = 0x07;


    string type_ndef_str = "";
    string record_name_str = "";
    string payload_str = "";
    int payload_size;

    byte[] payload_array;

    const int TYPE_LENGTH_INDEX = 1;
    const int PAYLOAD_LENGTH_INDEX = 2;

    string Card_URI_prefix = "";
    int TYPE_SP = 0x53;
    int TYPE_T = 0x54;
    int TYPE_U = 0x55;
    string Card_URI_content = "";
    string Card_URI_Tot = "";
    string Card_Title_lang = "";
    string Card_Title_content = "";
    bool Card_Title_present = false;
    int type_NDEF = 0;

    private string decodeTNF(byte tnf)
    {
      switch (tnf)
      {
        case EMPTY: return "Empty";
        case NFC_TYPE: return "NFC Forum well-known type";
        case MEDIA_TYPE: return "Media type as defined in RFC 2046";
        case ABSOLUTE_URI: return "Absolute URI as defined in RFC 3986";
        case EXTERNAL_TYPE: return "NFC Forum external type";
        case UNKNOWN_TYPE: return "Unknown";
        case UNCHANGED: return "Unchanged";
        case RESERVED: return "Reserved";
        default: return "Type unknown";
      }
    }

    NameValueCollection URI_Codes = new NameValueCollection();

    private void set_URI_Codes()
    {
      URI_Codes.Add("0x00", "");
      URI_Codes.Add("0x01", "http://www.");
      URI_Codes.Add("0x02", "https://www.");
      URI_Codes.Add("0x03", "http://");
      URI_Codes.Add("0x04", "https://");
      URI_Codes.Add("0x05", "tel:");
      URI_Codes.Add("0x06", "mailto:");
      URI_Codes.Add("0x07", "ftp://anonymous:anonymous@");
      URI_Codes.Add("0x08", "ftp://ftp.");
      URI_Codes.Add("0x09", "ftps://");
      URI_Codes.Add("0x0A", "sftp://");
      URI_Codes.Add("0x0B", "smb://");
      URI_Codes.Add("0x0C", "nfs://");
      URI_Codes.Add("0x0D", "ftp://");
      URI_Codes.Add("0x0E", "dav://");
      URI_Codes.Add("0x0F", "news:");
      URI_Codes.Add("0x10", "telnet://");
      URI_Codes.Add("0x11", "imap:");
      URI_Codes.Add("0x12", "rtsp://");
      URI_Codes.Add("0x13", "urn:");
      URI_Codes.Add("0x14", "pop:");
      URI_Codes.Add("0x15", "sip:");
      URI_Codes.Add("0x16", "sips:");
      URI_Codes.Add("0x17", "tftp:");
      URI_Codes.Add("0x18", "btspp://");
      URI_Codes.Add("0x19", "btl2cap://");
      URI_Codes.Add("0x1A", "btgoep://");
      URI_Codes.Add("0x1B", "tcpobex://");
      URI_Codes.Add("0x1C", "irdaobex://");
      URI_Codes.Add("0x1D", "file://");
      URI_Codes.Add("0x1E", "urn:epc:id:");
      URI_Codes.Add("0x1F", "urn:epc:tag:");
      URI_Codes.Add("0x20", "urn:epc:pat:");
      URI_Codes.Add("0x21", "urn:epc:raw:");
      URI_Codes.Add("0x22", "urn:epc:");
      URI_Codes.Add("0x23", "urn:nfc:");
      URI_Codes.Add("0xFE", "http://nfc.springcard.com/?");

    }

    private string URI_Identifier_Code(byte code)
    {
      return URI_Codes[byteToStringHex(code)];
    }

    private byte URI_Identifier_Code(string code)
    {
      byte value = 0x00;
      string value_str = "";
      foreach (string s in URI_Codes.AllKeys)
      {
        if (URI_Codes[s] == code)
          value_str = s;
      }
      if (value_str != "")
        value = StringToByte(value_str);
      return value;
    }

    private byte[] decodeNDEF(byte[] ndef)
    {
      int rest_index = 0;
      if (ndef.Length > 3)
      {
        int payload_length = 4;

        if ((ndef[0] & SHORT_RECORD_FLAG) != 0)
        {
          payload_length = 1;
          payload_size = (int)ndef[PAYLOAD_LENGTH_INDEX];
        }
        else
        {
          payload_size = (int)(
            ndef[PAYLOAD_LENGTH_INDEX] << 24
            + ndef[PAYLOAD_LENGTH_INDEX + 1] << 16
            + ndef[PAYLOAD_LENGTH_INDEX + 2] << 8
            + ndef[PAYLOAD_LENGTH_INDEX + 3]);
        }
        int type_length = (int)ndef[TYPE_LENGTH_INDEX];
        int type_index = PAYLOAD_LENGTH_INDEX + payload_length;
        int payload_index = type_index + type_length;
        if ((ndef[0] & ID_LENGTH_FLAG) != 0)
        {
          type_index++;
          payload_index += (int)ndef[PAYLOAD_LENGTH_INDEX + payload_length];
        }
        byte tnf = (byte)(ndef[0] & 0x07);
        type_ndef_str = decodeTNF(tnf);
        byte[] type_array = new byte[type_length];
        for (int cpt = 0; cpt < type_length; cpt++)
        {
          if (cpt + type_index < ndef.Length)
            type_array[cpt] = ndef[cpt + type_index];
        }
        record_name_str = createAsciiStrFromByteArray(type_array);
        rest_index = payload_index;
        if (payload_size > 0)
        {
          /* store payload in an array */
          payload_array = new byte[payload_size];
          for (int cpt = 0; cpt < payload_size; cpt++)
          {
            if (cpt + payload_index < ndef.Length)
              payload_array[cpt] = ndef[cpt + payload_index];
          }
          if (record_name_str == "Sp")
          {
            decodePayloadSp(payload_array);
            type_NDEF = TYPE_SP;
          }
          if (record_name_str == "U")
          {
            decodePayloadURI(payload_array);
            if (type_NDEF == 0)
            {
              type_NDEF = TYPE_U;
            }
          }
          if (record_name_str == "T")
          {
            decodePayloadText(payload_array);
            if (type_NDEF == 0)
            {
              type_NDEF = TYPE_T;
            }
          }

          /*payload_str = URI_Identifier_Code(ndef[payload_index]) +
            createAsciiStrFromByteArray(payload_array);*/
          rest_index = payload_index + payload_size;
        }
      }

      int rest_size = ndef.Length - rest_index;
      byte[] rest_array = new byte[0];
      if (rest_size > 0)
      {
        rest_array = new byte[rest_size];
        for (int cpt = 0; cpt < rest_size; cpt++)
        {
          if (cpt + rest_index < ndef.Length)
            rest_array[cpt] = ndef[cpt + rest_index];
        }
      }
      return rest_array;
    }




    #region decode Payload

    const int STATUS_INDEX = 0;
    const int LANGUAGE_CODE_INDEX = 1;

    private void decodePayloadText(byte[] payload)
    {
      Card_Title_present = true;
      /* first, decode status byte : retreive language code length */
      int langage_code_size = (int)(payload[STATUS_INDEX] & 0x3F);

      /* retreive language code */
      byte[] language_array = new byte[langage_code_size];
      for (int i = 0; i < langage_code_size; i++)
      {
        if (LANGUAGE_CODE_INDEX + i < payload.Length)
          language_array[i] = payload[LANGUAGE_CODE_INDEX + i];
      }
      Card_Title_lang = createAsciiStrFromByteArray(language_array);

      /* retreive text*/
      int text_index = LANGUAGE_CODE_INDEX + langage_code_size;
      if (text_index < payload.Length)
      {
        byte[] text_array = new byte[payload.Length - text_index];
        for (int i = 0; i < text_array.Length; i++)
        {
          if (i + text_index < payload.Length)
            text_array[i] = payload[i + text_index];
        }

        Card_Title_content = createAsciiStrFromByteArray(text_array);

      }


    }

    private void decodePayloadURI(byte[] payload)
    {
      if (payload.Length > 1)
      {
        byte[] payload_array_tmp = new byte[payload.Length - 1];
        for (int cpt = 0; cpt < payload_array_tmp.Length; cpt++)
        {
          payload_array_tmp[cpt] = payload[cpt + 1];
        }
        Card_URI_prefix = URI_Identifier_Code(payload[0]);
        Card_URI_content = createAsciiStrFromByteArray(payload_array_tmp);
        Card_URI_Tot = Card_URI_prefix + Card_URI_content;
        payload_str = URI_Identifier_Code(payload[0]) +
          createAsciiStrFromByteArray(payload_array_tmp);
      }

    }


    private void decodePayloadSp(byte[] payload)
    {
      byte[] rest;
      do
      {
        rest = decodeNDEF(payload);
        payload = rest;
      } while (rest.Length > 0);
    }

    #endregion

#endregion

#region conversion from str to byte and vice versa

    private byte[] createByteArrayFromStr(string StrIn)
    {
      string Str = StrIn.Replace("0x", "").Replace(" ", "");

      if (Str.IndexOf(" ") != -1)
      {
        string nStr = "";
        for (int cpt = 0; cpt < Str.Length; cpt = cpt + 2)
        {
          nStr = nStr + Str.Substring(cpt, 2) + " ";
        }
        Str = nStr;
      }

      // command stored into byte array
      char[] delimiterByte = new char[1];
      delimiterByte[0] = ' ';
      string properStr = Str;
      // removing last " "
      if (Str.EndsWith(" "))
      {
        properStr = Str.Substring(0, Str.Length - 1);
      }

      string[] Strbuff = properStr.Split(delimiterByte);
      byte[] StrBuffer = new byte[Strbuff.Length];
      for (int i = 0; i < Strbuff.Length; i++)
      {
        // converting byte string to byte  
        StrBuffer[i] = System.Convert.ToByte(int.Parse(Strbuff[i],
          System.Globalization.NumberStyles.HexNumber));
      }
      return StrBuffer;
    }

    private string createStrFromByteArray(byte[] byteArray)
    {
      string Str = "";
      if (byteArray.Length <= 0)
        return Str;
      for (int cpt = 0; cpt < byteArray.Length - 1; cpt++)
        Str += String.Format("{0:X02}", byteArray[cpt]) + " ";
      Str += String.Format("{0:X02}", byteArray[byteArray.Length - 1]);
      return Str;
    }

    private byte[] createByteArrayFromAsciiStr(string StrIn)
    {
      byte[] array = new byte[StrIn.Length];

      for (int cpt = 0; cpt < StrIn.Length; cpt++)
        array[cpt] = (byte)System.Convert.ToChar(StrIn[cpt]);

      return array;

    }

    private byte StringToByte(string StrIn)
    {
      string Str = StrIn.Replace("0x", "").Replace(" ", "");
      return System.Convert.ToByte(int.Parse(Str,
          System.Globalization.NumberStyles.HexNumber));
    }

    private string createAsciiStrFromByteArray(byte[] byteArray)
    {
      string Str = "";
      if (byteArray.Length <= 0)
        return Str;
      for (int cpt = 0; cpt < byteArray.Length; cpt++)
        Str += byteToAscii(byteArray[cpt]);
      return Str;
    }

    public string byteToAscii(byte value)
    {
      if (value < 32 || value > 127)
        return (".");
      else
        return (System.Convert.ToChar(value).ToString());
    }

    public string byteToString(byte value)
    {
      if (value < 32 || value > 127)
        return (String.Format("{0:X02}", value));
      else
        return (System.Convert.ToChar(value).ToString());
    }

    public string byteToStringHex(byte value)
    {
      return ("0x" + String.Format("{0:X02}", value));
    }
#endregion

#region determine card type
    int pixnnIndex = 26;
    const int JEWEL = 1;
    const int ULTRALIGHT = 2;
    const int ULTRALIGHTC = 5;
    const int MIFARE_1K = 6;
    const int MIFARE_4K = 7;
    const int TYPE_UNSUPPORTED = 0;
    int card_type;

    private void GetCardType()
    {
      string AtrStr = scard.CardAtr.AsString("");
      card_type = TYPE_UNSUPPORTED;
      if (AtrStr.Length > pixnnIndex + 4)
      {
        if (AtrStr.Substring(pixnnIndex, 4) == "0003")
        {
          card_type = ULTRALIGHT;
          labelInfo.Text = "Ultralight card found, reading data...";
        }
        else
        {
          if (AtrStr.Substring(pixnnIndex, 4) == "0030")
          {
            card_type = JEWEL;
            labelInfo.Text = "Jewel card found, reading data...";
          }
          else
          {
            if (AtrStr.Substring(pixnnIndex, 4) == "0001")
            {
              card_type = MIFARE_1K;
              labelInfo.Text = "Mifare 1K card found, reading data...";
            }
            else
            {
              if (AtrStr.Substring(pixnnIndex, 4) == "0002")
              {
                card_type = MIFARE_4K;
                labelInfo.Text = "Mifare 4K card found, reading data...";
              }
              else
              {
                labelInfo.Text = "Unsupported card type";
              }
            }
          }
        }
      }else
        labelInfo.Text = "Unsupported card type";
    }
#endregion

#region open URL
     [System.Runtime.InteropServices.DllImportAttribute("shell32.dll", SetLastError = true)]
     static extern int ShellExecute(int hwnd, string lpOperation, string lpFile, string lpParameters, string lpDirectory, int nShowCmd);

    public void RunUrl(string sUrl, bool bMaximized)
    {
        ShellExecute(0, "open", sUrl, string.Empty, string.Empty, (bMaximized == true) ? 3 : 1);
    }


#endregion

#region readCard

    private void ReadCard()
    {
      if (card_type == JEWEL)
        ReadJewel();
      if (card_type == ULTRALIGHT)
        ReadMifareUL();
      if (card_type == MIFARE_1K || card_type == MIFARE_4K)
        ReadMifareClassic();

    }

    #region mifare classic
    private const int SZ_SECTOR_MIF1K = 48;
    private const byte AID_1 = 0x03;
    private const byte AID_2 = 0xE1;

    bool NFC_Content = false;
    byte blockToRead = 0x00;

    byte[] card_data;
    int offset_card;


    int nb_sectors_nfc = 0;
    int nb_sectors_read = 0;

    void ReadMifareClassic()
    {
      Cursor = Cursors.WaitCursor;


      nb_sectors_read = 0;
      CAPDU capdu1 = new CAPDU("FF F3 00 01 20"); //reading first block
      scard.Transmit(capdu1, new SCardChannel.TransmitDoneCallback(ReadMifareBlock));
    }



    delegate void TransmitDoneInvoker(RAPDU rapdu);

    void ReadMifareBlock(RAPDU rapdu)
    {
      if (this.InvokeRequired)
      {
        this.BeginInvoke(new TransmitDoneInvoker(ReadMifareBlock), rapdu);
        return;
      }

    if (rapdu == null)
    {
      /* An error has occured */
      labelInfo.Text = "Error " + scard.LastErrorAsString;
    }
    else
    {
      if (blockToRead == 0)
      {

        for (int i = 2; i < rapdu.data.Bytes.Length; i = i + 2)
        {
          if ((rapdu.data.Bytes[i] == AID_1) && (rapdu.data.Bytes[i + 1] == AID_2))
            nb_sectors_nfc++;
        }
        if (nb_sectors_nfc > 0)
        {
          NFC_Content = true;
          card_data = new byte[nb_sectors_nfc * SZ_SECTOR_MIF1K];
          offset_card = 0;
        }
        else
        {
          labelInfo.Text = "No NFC Content found on this card";
          Cursor = Cursors.Default;
        }
      }
      if (rapdu.SW1 == 0x90 && rapdu.SW2 == 0x00)
      {
        if (NFC_Content && blockToRead > 0)
        {
          for (int i = 0; i < rapdu.data.Bytes.Length; i++)
          {
            card_data[offset_card + i] = rapdu.data.Bytes[i];
          }
        }
        if (blockToRead > 0)
          offset_card += rapdu.data.Bytes.Length;
        blockToRead = (byte)(blockToRead + 4);
        nb_sectors_read++;
        if (nb_sectors_nfc == (nb_sectors_read - 1) && NFC_Content)
          blockToRead = 0x40;
        if (!NFC_Content)
          blockToRead = 0x40;

        if (blockToRead <= 0x3C)
        {
          CAPDU capdu2 = new CAPDU("FF F3 00 " + String.Format("{0:X02}", blockToRead) + " 00");
          scard.Transmit(capdu2, new SCardChannel.TransmitDoneCallback(ReadMifareBlock));
        }
      }else {
        labelInfo.Text = "Error while reading this card - Try another card";
        Cursor = Cursors.Default;
      }

    }
  

      if (NFC_Content && blockToRead == 0x40)
      {
        decodeTLV(card_data);
        type_NDEF = 0;
        decodeNDEF(V_VALUE);
        if (Card_URI_Tot != "")
        {
          panelPrompt.Visible = true;
          labelQuestion.Text = "Would you like to open " + Card_URI_Tot + " ?";
          if (alwaysOpen)
          {
            RunUrl(Card_URI_Tot, true);
            labelInfo.Text = "Opening URL " + Card_URI_Tot + " , please wait ...";
          }
          if (neverOpen)
            labelQuestion.Text = "URL found on card : " + Card_URI_Tot;
        }
        Cursor = Cursors.Default;
        return;
      }

      cpt_sector++;


    }


    #endregion

    #region ultralight
    const int MAX_NDEF_UL = 46;
    const int MAX_NDEF_ULC = 142;
    byte page;
    int offset_cardContent;

    private void ReadMifareUL()
    {
      offset_cardContent = 0;
      card_data = new byte[192];
      card_type = ULTRALIGHT;
      page = 4;

      Cursor = Cursors.WaitCursor;


      nb_sectors_read = 0;
      CAPDU capdu1 = new CAPDU("FF B0 00 04 10"); //reading first block
      scard.Transmit(capdu1, new SCardChannel.TransmitDoneCallback(ReadUltralightPage));

    }

    delegate void TransmitUlDoneInvoker(RAPDU rapdu);

    void ReadUltralightPage(RAPDU rapdu)
    {
      if (this.InvokeRequired)
      {
        this.BeginInvoke(new TransmitUlDoneInvoker(ReadUltralightPage), rapdu);
        return;
      }

      if (rapdu == null)
      {
        /* An error has occured */
        labelInfo.Text = "Error " + scard.LastErrorAsString;
      }
      else
      {
        if (rapdu.SW1 == 0x90 && rapdu.SW2 == 0x00)
        {
          for (int i = 0; i < rapdu.data.Bytes.Length; i++)
          {
            if (offset_cardContent + i < card_data.Length)
              card_data[offset_cardContent + i] = rapdu.data.Bytes[i];
          }
          page = (byte)(page + 4);
          offset_cardContent += 16;
          if ((page > 16))
            card_type = ULTRALIGHTC;

          if (page < 44)
          {
            CAPDU capdu2 = new CAPDU("FF B0 00 " + String.Format("{0:X02}", page) + " 10");
            scard.Transmit(capdu2, new SCardChannel.TransmitDoneCallback(ReadUltralightPage));
          }
        }
        else
        {
          page = 44;
          Cursor = Cursors.Default;
        }

      }


      if (page == 44)
      {
        decodeTLV(card_data);
        type_NDEF = 0;
        decodeNDEF(V_VALUE);
        if (Card_URI_Tot != "")
        {
          panelPrompt.Visible = true;
          labelQuestion.Text = "Would you like to open " + Card_URI_Tot + " ?";
          if (alwaysOpen)
          {
            RunUrl(Card_URI_Tot, true);
            labelInfo.Text = "Opening URL " + Card_URI_Tot + " , please wait ...";
          }
          if (neverOpen)
            labelQuestion.Text = "URL found on card : " + Card_URI_Tot;
        }
        else
        {
          labelInfo.Text = "No NFC Content found on this card";
        }
        Cursor = Cursors.Default;
        return;
      }


    }

    #endregion

    #region Jewel


    private void ReadJewel()
    {
      Cursor = Cursors.WaitCursor;
      card_data = new byte[108];

      
      CAPDU capdu1 = new CAPDU("FF B0 00 0C 6C"); //reading whole card
      scard.Transmit(capdu1, new SCardChannel.TransmitDoneCallback(ReadJewelDone));
    
    }

    delegate void TransmiJewelDoneInvoker(RAPDU rapdu);

    void ReadJewelDone(RAPDU rapdu)
    {
      if (this.InvokeRequired)
      {
        this.BeginInvoke(new TransmitUlDoneInvoker(ReadJewelDone), rapdu);
        return;
      }

      if (rapdu == null)
      {
        /* An error has occured */
        labelInfo.Text = "Error " + scard.LastErrorAsString;
      }
      else
      {
        if (rapdu.SW1 == 0x90 && rapdu.SW2 == 0x00)
        {
          for (int i = 0; i < rapdu.data.Bytes.Length; i++)
          {
              card_data[i] = rapdu.data.Bytes[i];
          }
        }
        else
        {
          labelInfo.Text = "Cannot read Jewel card";
          Cursor = Cursors.Default;
        }

      }


      decodeTLV(card_data);
      type_NDEF = 0;
      decodeNDEF(V_VALUE);
      if (Card_URI_Tot != "")
      {
        panelPrompt.Visible = true;
        labelQuestion.Text = "Would you like to open " + Card_URI_Tot + " ?";
        if (alwaysOpen)
        {
          RunUrl(Card_URI_Tot, true);
          labelInfo.Text = "Opening URL " + Card_URI_Tot + " , please wait ...";
        }
        if (neverOpen)
          labelQuestion.Text = "URL found on card : " + Card_URI_Tot;
          
      }
      else
      {
        labelInfo.Text = "No NFC Content found on this card";
      }
      Cursor = Cursors.Default;


    }

    #endregion



#endregion





  }
}