using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.IO;
using System.Windows.Forms;
using System.Collections.ObjectModel;
using System.Runtime.InteropServices;

using SpringCardPCSC;


namespace csScriptor
{
  public partial class MainForm : Form
  {
    const int SCARD_PROTOCOL_UNDEFINED = 0;
    //const int SCARD_S_SUCCESS = 0;
    const int SCARD_RESET_CARD = 0;
    const int SCARD_PROTOCOL_T0 = 1;
    const int SCARD_PROTOCOL_T1 = 2;
    const int SCARD_PROTOCOL_RAW = 0x00010000;

    private IntPtr SCARD_PCI_T0 = IntPtr.Zero;
    private IntPtr SCARD_PCI_T1 = IntPtr.Zero;
    private IntPtr SCARD_IO_REQUEST = IntPtr.Zero;

    private string format = "{0:X02}";

    private int nContext;		//card reader context handle
    private int nCard = 0;			//card connection handle
    public int nActiveProtocol = SCARD_PROTOCOL_T0 | SCARD_PROTOCOL_T1;
    public string nSelectedReader;              //Reader to use
    public string ScriptFileName = "";

    //Smart Card PCSC dll functions (windows)
    [DllImport("WinScard.dll", EntryPoint = "SCardEstablishContext")]
    public static extern uint ScriptorEstablishContext(uint dwScope,int nNotUsed1,
      int nNotUsed2, ref int phContext);
    [DllImport("WinScard.dll", EntryPoint = "SCardConnect")]
    public static extern uint ScriptorConnect(int hContext, string cReaderName,
      uint dwShareMode, uint dwPrefProtocol, ref int phCard, 
      ref int ActiveProtocol);
    [DllImport("WinScard.dll", EntryPoint = "SCardReconnect")]
    public static extern uint ScriptorReconnect(int hCard,uint dwShareMode, 
      uint dwPrefProtocol, uint swInit, ref int ActiveProtocol);
    [DllImport("WinScard.dll", EntryPoint = "SCardDisconnect")]
    public static extern uint ScriptorDisconnect(int hCard, int Disposition);
    [DllImport("winscard.dll", EntryPoint = "SCardListReaders", 
      SetLastError = true, CharSet = CharSet.Unicode)]
		internal static extern int ScriptorListReaders(int context, string groups,
      string readers, ref int size);
    [DllImport("winscard.dll", EntryPoint = "SCardTransmit", SetLastError = true)]
    internal static extern uint ScriptorTransmit(int hCard,
      IntPtr pioSendPci, byte[] pbSendBuffer,
      uint cbSendLength, IntPtr pioRecvPci, [In,Out] byte[] pbRecvBuffer,
      [In,Out] ref uint pcbRecvLength);
    [DllImport("winscard.dll", EntryPoint = "SCardGetStatusChange", CharSet = CharSet.Unicode)]
    public static extern uint ScriptorGetStatusChange(int hContext, int dwTimeout,
      [In, Out, MarshalAs(UnmanagedType.LPArray, SizeParamIndex =3)] 
      SCARD_READERSTATE[] rgReaderState, uint cReaders);

 
    public MainForm()
    {
      //SplashForm MySplashForm = new SplashForm();
      //MySplashForm.Show();

      InitializeComponent();
          
      tSStatusLabel.Text = "Welcome";

      //First step in using smart cards is CSardEstablishContext()
      uint nContext = 2;		//system
      int nNotUsed1 = 0;
      int nNotUsed2 = 0;
      this.nContext = 0;		//handle to context for future calls to use

      uint RetEstablish = ScriptorEstablishContext(nContext, nNotUsed1, nNotUsed2,
        ref this.nContext);
      if (RetEstablish != SCARD_S_SUCCESS)
      {
        tSStatusLabel.Text = 
          "error occured while calling ScriptorEstablishContext() "+
           codeToStr(RetEstablish);
        return;
      }
      ThStatusDeleg = new ThreadFinishedDelegate(onStatusDone);
      string[] cReaders = getListReaders();
      
      if (cReaders.Length > 0)
      {
        tSStatusLabel.Text = "Reader: " + cReaders[0] +
          ". Hit Ctl+C to connect";
        this.nSelectedReader = cReaders[0];
      }
    }

    


    private void connectReader(object sender, EventArgs e)
    {
      uint nShareMode = 1;			//exclusive
      uint nPrefProtocol = (uint) this.nActiveProtocol;	
      this.nCard = 0;
      uint RetConnect = ScriptorConnect(this.nContext, this.nSelectedReader, 
        nShareMode, nPrefProtocol, ref this.nCard, ref this.nActiveProtocol);
      if (RetConnect != SCARD_S_SUCCESS)
        tSStatusLabel.Text = "error occured while calling ScriptorConnect() : "+
          codeToStr(RetConnect);
      else
        tSStatusLabel.Text = this.nSelectedReader+" connected";
    }

    public void DisconnectReader(object sender, EventArgs e)
    {
      Disconnect();
    }

    public void Disconnect()
    {
      int nDisposition = 0;		//don't do power down, eject, etc...
      if (this.nCard != 0)
      {
        uint RetDisconnect = ScriptorDisconnect(this.nCard, nDisposition);
        this.nCard = 0;
        if (RetDisconnect != SCARD_S_SUCCESS)
          tSStatusLabel.Text =
            "error occured while calling ScriptorDisconnect() : " +
            codeToStr(RetDisconnect);
        else
        {
          tSStatusLabel.Text = "Disconnected ";
        }
      }
      else
        tSStatusLabel.Text = "not connected ";
    }

    private void SetReader(object sender, EventArgs e)
    {
      SettingsForm nSettings = new SettingsForm(this);
      nSettings.ShowDialog();
      this.nSelectedReader = nSettings.getReader();
      this.nActiveProtocol = nSettings.getProtocol();
      tSStatusLabel.Text =  this.nSelectedReader +" selected."+
        " Hit Ctl+C to connect.";
    }

    public string[] getListReaders()
    {
      int size = 2048;
      int nbReaders = 0; 
      Collection<string> readers = new Collection<string>();
      if (ScriptorListReaders(this.nContext, null, null, ref size) == SCARD_S_SUCCESS)
      {
        // Allocate a string of the proper size.
        string names = new string(' ', size);
        if (ScriptorListReaders(this.nContext, null, names, ref size) ==
          SCARD_S_SUCCESS)
        {
          // The 'names' string will contain a multi-string of the,
          // reader names i.e. they are separated by 0x00 characters.
          string name = string.Empty;

          for (int i = 0; i < names.Length; i++)
          {
            if (names[i] == 0x00)
            {
              if (name.Length > 0)
              {
                //
                // We have the reader name.
                //
                readers.Add(name);
                name = string.Empty;
                nbReaders++;
              }
            }
            else
            {
              // Append found character.
              name = name + new string(names[i], 1);
            }
          }
        }
        else
        {
          tSStatusLabel.Text = "Cannot get Readers List ";
        }
      }
      else
      {
        tSStatusLabel.Text = "Cannot get Readers List ";
      }
      
      string[] cReaders = new string[nbReaders];
      readers.CopyTo(cReaders, 0);
      
      return cReaders;
    }

    private void ClearResult(object sender, EventArgs e)
    {
      richTextBoxResult.Clear();
    }
    
    private void ReconnectReader(object sender, EventArgs e)
    {
      if (this.nCard != 0)
      {
        uint nShareMode = 1;
        uint nPrefProtocol = (uint)this.nActiveProtocol;
        uint swInit = SCARD_RESET_CARD; 
        uint RetReconnect = ScriptorReconnect(this.nCard, nShareMode, nPrefProtocol,
          swInit, ref this.nActiveProtocol);
        if (RetReconnect != SCARD_S_SUCCESS)
        {
          tSStatusLabel.Text =
         "error occured while calling ScriptorReconnect() "+
         codeToStr(RetReconnect);
        }
        else
        {
          tSStatusLabel.Text = this.nSelectedReader+" reconnected";
        }
      }
      else
      {
        tSStatusLabel.Text = "Not connected. Hit Ctl+C to connect";
      }
    }

    private void RunScript(object sender, EventArgs e)
    {
      Run();
    }

    private void runSMenuItem_Click(object sender, EventArgs e)
    {
      Run();
    }

    public void Run()
    {

      if (this.nCard != 0)
      {
        // getting user entry
        string Script = richTextBoxScript.Text;

        char[] delimiter = new char[1];
        delimiter[0] = '\n';
        // seperating apdus (delimiter = '\n')
        string[] lines = Script.Split(delimiter);
        richTextBoxResult.SelectionColor = Color.Black;
        richTextBoxResult.AppendText("\n\nBeginning Script Execution...");
        string cmd = "";
        foreach (string line in lines)
        {
          // skip comment
          if (line.StartsWith("#"))
            continue;

          // skip blank lines
          string apdu = line.Replace(" ","");
          if (apdu.Equals(""))
            continue;

          string napdu = "";

          // if there isn't any space in the command, 
          // we add spaces every 2 character
          if (!apdu.Contains(" "))
          {
            for (int cpt = 0; cpt < apdu.Length; cpt = cpt + 2)
            {
              napdu = napdu + apdu.Substring(cpt, 2) + " ";
            }
            apdu = napdu;
          }

          // if line ends with \, concatenation with next line
          if (apdu.EndsWith("\\"))
          {
            // remove last character
            napdu = apdu.Remove(apdu.Length - 1);
            cmd = cmd + napdu + " ";
            continue;
          }

          cmd = cmd + apdu;


          // command stored into byte array
          char[] delimiterByte = new char[1];
          delimiterByte[0] = ' ';
          string properCmd =cmd;
          // removing last " "
          if (cmd.EndsWith(" "))
          {
            properCmd = cmd.Substring(0,cmd.Length-1);
          }


          if (cmd.Replace(" ","").Length % 2 != 0)
          {// wrong cmd size 
            richTextBoxResult.SelectionColor = Color.Red;
            richTextBoxResult.AppendText("\nWrong APDU Length");
            return;
          }

          
          
          string[] cmdbuff = properCmd.Split(delimiterByte);
          byte[] ApduBuffer = new byte[cmdbuff.Length];
          for (int i = 0; i < cmdbuff.Length; i++)
          {
            // converting byte string to byte  
            try
            {
              ApduBuffer[i] = System.Convert.ToByte(int.Parse(cmdbuff[i],
            System.Globalization.NumberStyles.HexNumber));
              
            }
            catch (Exception) 
            {
              richTextBoxResult.SelectionColor = Color.Red;
              richTextBoxResult.AppendText("\nCannot convert APDU");
              return;
            }
          }
          richTextBoxResult.SelectionColor = Color.Black;
          richTextBoxResult.AppendText("\nSending : ");


          richTextBoxResult.SelectionColor = Color.Blue;
          richTextBoxResult.AppendText(cmd);

          byte[] ApduResponse = new byte[260];
          uint ResponseLength = (uint)ApduResponse.Length;

          uint RetTransmit = ScriptorTransmit(this.nCard, SCARD_IO_REQUEST,
            ApduBuffer, (uint)ApduBuffer.Length, IntPtr.Zero, ApduResponse,
            ref ResponseLength);
          cmd = "";
          string answer = "";
          if (RetTransmit != SCARD_S_SUCCESS)
            answer = "error during ScriptorTransmit() " + codeToStr(RetTransmit);
          else
          {
            answer = showResp(ApduResponse, ResponseLength);
          }

          richTextBoxResult.SelectionColor = Color.Black;
          richTextBoxResult.AppendText("\nReceived : ");

          richTextBoxResult.SelectionColor = Color.Red;
          richTextBoxResult.AppendText(answer);
          richTextBoxResult.AppendText("\n");


        }
      }
    }

    private void openMenuItem_Click(object sender, EventArgs e)
    {
      openFileDialog.ShowDialog();
    }

    private void ReadScriptFile(object sender, CancelEventArgs e)
    {
      tSStatusLabel.Text= "User file opened";
      Stream userFileStream = new FileStream(openFileDialog.FileName, 
        FileMode.Open, FileAccess.Read);
      StreamReader userFileStreamReader = new StreamReader(userFileStream);
      string userFileString = userFileStreamReader.ReadToEnd();
      richTextBoxScript.Text = userFileString;
      ScriptFileName = openFileDialog.FileName;
      
      userFileStreamReader.Close();
      userFileStream.Close();
    }

    private void saveFileDialog_FileOk(object sender, CancelEventArgs e)
    {
      tSStatusLabel.Text = "Saving Script";
      ScriptFileName = saveFileDialog.FileName;
      writeFile();
    }

    private void saveAsMenuItem_Click(object sender, EventArgs e)
    {
      string Script = richTextBoxScript.Text;
      // is there anything to save?
      if (Script != "")
      {
        saveFileDialog.ShowDialog();
      }else
      {
        tSStatusLabel.Text = "Nothing to save";
      }
    }

    private void saveMenuItem_Click(object sender, EventArgs e)
    {
      string Script = richTextBoxScript.Text; 
      // is there anything to save?
      if (Script != "")
      {
        if (ScriptFileName == "")
        {//no default file
          saveFileDialog.ShowDialog();
        }else 
        {// writing on default file 
          writeFile();
        }
      }
      else
      {
        tSStatusLabel.Text = "Nothing to save";
      }
    }

    public void writeFile()
    {
      // getting user entry
      string Script = richTextBoxScript.Text;

      char[] delimiter = new char[1];
      delimiter[0] = '\n';
      // seperating apdus (delimiter = '\n')
      string[] lines = Script.Split(delimiter);

      Stream writeFileStream = new FileStream(ScriptFileName,
        FileMode.OpenOrCreate, FileAccess.Write);
      StreamWriter writeFileStreamWriter = new StreamWriter(writeFileStream);
      
      
      foreach (string line in lines) 
      {
        writeFileStreamWriter.WriteLine(line);
      }
      
      writeFileStreamWriter.Close();
      writeFileStream.Close();
      tSStatusLabel.Text = "Script saved";
      
    }

    private void NewScriptFile(object sender, EventArgs e)
    {
      // erasing default file name
      ScriptFileName = "";

      // erasing script area
      richTextBoxScript.Text = "";

      tSStatusLabel.Text = "";
    }

    private void checkBoxWrap_CheckedChanged(object sender, EventArgs e)
    {
      richTextBoxScript.WordWrap = checkBoxWrap.Checked;
    }



    private void quitMenuItem_Click(object sender, EventArgs e)
    {
      Close();
    }
    private bool AskConfirmQuitAppli()
    {
      if (MessageBox.Show("Are you sure you want to quit?",
      "Confirmation",
      MessageBoxButtons.OKCancel) == DialogResult.OK)
        return true;
      return false;
    }

    private void MainForm_FormClosing(object sender, FormClosingEventArgs e)
    {
      if (AskConfirmQuitAppli() == false)
        e.Cancel = true;
    }

    private void MainForm_FormClosed(object sender, FormClosedEventArgs e)
    {
      // disconnect the reader before closing application
      Disconnect();
      timerStatus.Stop();
    }

    private void timer1_Tick(object sender, EventArgs e)
    {
      CheckStatus();
    }

    /*
     * Structure used in GetStatusChange
     */

    [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Auto)]
    public struct SCARD_READERSTATE
    {
      internal string szReader;
      internal IntPtr pvUserData;
      internal uint dwCurrentState;
      internal uint dwEventState;
      internal uint cbAtr;
      [MarshalAs(UnmanagedType.ByValArray, SizeConst = 0x24,
      ArraySubType = UnmanagedType.U1)]
      internal byte[] rgbAtr;
    }
    
    
    
    public string CheckStatus()
    {
      // Getting card status
      int TimeoutInf = 0; //infinite timeout
      uint numreaders = 1;
      SCARD_READERSTATE[] states = new SCARD_READERSTATE[1];
      states[0] = new SCARD_READERSTATE();
      states[0].szReader = this.nSelectedReader;
      states[0].pvUserData = IntPtr.Zero;
      states[0].dwCurrentState = 0;
      states[0].dwEventState = 0;
      states[0].cbAtr = 0;
      states[0].rgbAtr = null;
      uint retStatusChange = 0;

      retStatusChange = ScriptorGetStatusChange(this.nContext, TimeoutInf,
        states, numreaders);

      if (retStatusChange != SCARD_S_SUCCESS)
      {
        tSStatusLabel.Text = "Error during ScriptorGetStatusChange()"+
          codeToStr(retStatusChange);
      }
      return StateToStr(states[0].dwEventState);
    }

    private void aboutCScriptorToolStripMenuItem_Click(object sender, EventArgs e)
    {
      AboutForm aboutFrm = new AboutForm();
      aboutFrm.ShowDialog();
    }

    private void MainForm_KeyDown(object sender, KeyEventArgs e)
    {
      if (e.KeyCode == Keys.F5)
        Run();
    }

    private void radioButtonAscii_CheckedChanged(object sender, EventArgs e)
    {
      if (radioButtonAscii.Checked)
        format = "asc";
    }

    private void butClearRes_Click(object sender, EventArgs e)
    {
      richTextBoxResult.Clear();
    }

#region status

    System.Threading.Thread StatusThread = null; /* Get reader status */
    public ThreadFinishedDelegate ThStatusDeleg;
    public delegate void ThreadFinishedDelegate();
    public string StatusStr = "";

    public void ThStatusProc()
    {
      string str = CheckStatus();
      StatusStr = str;
      if (str.Contains(":"))
        StatusStr = StatusStr.Remove(str.IndexOf(':'));
      Invoke(ThStatusDeleg);
    }

    public void onStatusDone()
    {
      CardStatustextBox.Text = StatusStr;

      if (CardStatustextBox.Text == "")
      {
        pictureBoxAbs.Hide();
        pictureBoxAl.Hide();
        pictureBoxC.Hide();
        pictureBoxP.Hide();
        buttonConnect.Enabled = false;
        buttonRun.Enabled= false;
      }
      if (CardStatustextBox.Text.Contains("PRESENT"))
      {
        pictureBoxAbs.Hide();
        pictureBoxAl.Hide();
        pictureBoxC.Hide();
        pictureBoxP.Show();
        buttonRun.Enabled = false;
        buttonConnect.Enabled = true;
        // buttonDisco.Hide();
        // buttonConnect.Text = "Connect";
      }
      if (CardStatustextBox.Text.Contains("INUSE"))
      {
        pictureBoxAbs.Hide();
        pictureBoxAl.Hide();
        pictureBoxC.Show();
        pictureBoxP.Hide();
        buttonRun.Enabled = true;
        //   buttonDisco.Show();
        // buttonConnect.Text = "Reconnect";
      }
      if (CardStatustextBox.Text.Contains("EMPTY"))
      {
        pictureBoxAbs.Show();
        pictureBoxAl.Hide();
        pictureBoxC.Hide();
        pictureBoxP.Hide();
        buttonRun.Enabled = false;
        buttonConnect.Enabled = false;
        //  buttonDisco.Hide();
        // buttonConnect.Text = "Connect";
      }

    }

    private void timerStatus_Tick(object sender, EventArgs e)
    {
      StatusThread = new System.Threading.Thread(new
      System.Threading.ThreadStart(ThStatusProc));
      StatusThread.Start();
    }
#endregion

    private void buttonConnect_Click(object sender, EventArgs e)
    {
      uint nShareMode = 1;			//exclusive
      uint nPrefProtocol = (uint)this.nActiveProtocol;
      this.nCard = 0;
      uint RetConnect = ScriptorConnect(this.nContext, this.nSelectedReader,
        nShareMode, nPrefProtocol, ref this.nCard, ref this.nActiveProtocol);
      if (RetConnect != SCARD_S_SUCCESS)
        tSStatusLabel.Text = "error occured while calling ScriptorConnect() : " +
          codeToStr(RetConnect);
      else
        tSStatusLabel.Text = this.nSelectedReader + " connected"; 
    }

    private void radioButtonBoth_CheckedChanged(object sender, EventArgs e)
    {
      if (radioButtonBoth.Checked)
        format = "mix";
    }

    private void radioButtonHex_CheckedChanged(object sender, EventArgs e)
    {
      if(radioButtonHex.Checked)
        format = "{0:X02}";
    }


    private string showResp(byte[] resp, uint length)
    {
      string answer = "";
      if (format != "mix")
      {
        for (int cpt = 0; cpt < length; cpt++)
        {
          if (format == "{0:X02}")
            answer = answer + String.Format(format, resp[cpt])
            + " ";
          else
          {
            if (format == "asc")
            {
              if ((resp[cpt] > 31) && (resp[cpt] < 128))
                answer = answer +
                    System.Convert.ToChar(resp[cpt]) + " ";
              else
                answer = answer + " ";
            }
          }
        }
      }
      else
      {
        int col=0;
        answer = "\n";
        string hex = "", asc="";
        for (int cpt = 0; cpt < length; cpt++)
        {
          if (col == 16)
          {
            answer += hex + "\t|| " + asc+ "\n";
            col = 0;
            hex = asc = "";
          }

          hex = hex + String.Format("{0:X02}", resp[cpt])
            + "  ";
     
           
          if ((resp[cpt] > 31) && (resp[cpt] < 128))
            asc = asc+
                System.Convert.ToChar(resp[cpt]) + "";
          else
            asc = asc + "";

          col++;
        }
        answer += hex + "\t|| " + asc;
      }
      return (answer);
    }

    private void richTextBoxResult_TextChanged(object sender, EventArgs e)
    {
      // scroll bar
      // Put the cursor at the very end of the text
      richTextBoxResult.Select((richTextBoxResult.Text.Length), 0);

      // Scroll
      richTextBoxResult.ScrollToCaret();

    }

    private void richTextBoxScript_TextChanged(object sender, EventArgs e)
    {
      // scroll bar
      // Put the cursor at the very end of the text
      richTextBoxScript.Select((richTextBoxScript.Text.Length), 0);

      // Scroll
      richTextBoxScript.ScrollToCaret();

    }
  }
}
