﻿/*
 * Crée par SharpDevelop.
 * Utilisateur: johann
 * Date: 20/04/2010
 * Heure: 15:42
 * 
 * Pour changer ce modèle utiliser Outils | Options | Codage | Editer les en-têtes standards.
 */
using System;
using System.Collections.Generic;
using System.Drawing;
using System.Windows.Forms;

using SpringCardPCSC;

namespace spcsc_net
{
	public partial class MainForm : Form
	{
		SCardReader reader = null;
		SCardChannel scard = null;
		
		public MainForm()
		{
			InitializeComponent();
		}
		
		delegate void ReaderStatusChangedInvoker(uint ReaderState, CardBuffer CardAtr);
		
		void OnReaderStatusChanged(uint ReaderState, CardBuffer CardAtr)
		{
			if (this.InvokeRequired)
			{
				this.BeginInvoke(new ReaderStatusChangedInvoker(OnReaderStatusChanged), ReaderState, CardAtr);
				return;
			}
			
			eStatus.Text = SCARD.ReaderStatusToString(ReaderState);
			
			if ((ReaderState & SCARD.STATE_PRESENT) != 0)
			{
				if (CardAtr != null)
					eCardATR.Text = CardAtr.AsString(" ");
				else
					eCardATR.Text = "";
				
				if (scard == null)
				{
					btnConnect.Enabled = true;
				}

			} else
			{
				if (scard != null)
				{
					if (scard.Connected)
						scard.Disconnect();

					scard = null;
				}

				btnConnect.Enabled = false;
				btnTransmit.Enabled = false;
				btnDisconnect.Enabled = false;

				eCardATR.Text = "";
				
				if (ReaderState == 0)
				{
					/* Unallowed value - an error has occured */
					StatusInfo.Text = reader.LastErrorAsString;
				}
			}
		}
		
		void MainFormShown(object sender, EventArgs e)
		{
			BtnRefreshListClick(sender, e);
		}	
		
		void BtnRefreshListClick(object sender, EventArgs e)
		{
			cbReaders.Items.Clear();
			
			Cursor = Cursors.WaitCursor;
		
			for (int i=0; i<SCARD.Readers.Length; i++)
				cbReaders.Items.Add(SCARD.Readers[i]);
			
			if (cbReaders.Items.Count > 0)
				cbReaders.SelectedIndex = 0;
			
			Cursor = Cursors.Default;
		}
		
		void BtnConnectClick(object sender, EventArgs e)
		{
			scard = new SCardChannel(cbReaders.Text);

            if (cbProtocol.Text == "")
                cbProtocol.SelectedIndex = 0;
			
			scard.ProtocolAsString = cbProtocol.Text;
			scard.ShareModeAsString = "SHARED";
			
			scard.Connect();
			
			if (!scard.Connected)
			{
				/* An error has occured */
				StatusInfo.Text = scard.LastErrorAsString;
			} else
			{
				/* OK */
				StatusInfo.Text = "Connected";
				
				eCardATR.Text = scard.CardAtr.AsString(" ");
			    cbProtocol.Text = scard.ProtocolAsString;
				
				btnConnect.Enabled = false;
				btnTransmit.Enabled = true;
				btnDisconnect.Enabled = true;
			}
		}
		
		void BtnDisconnectClick(object sender, EventArgs e)
		{
			StatusInfo.Text = "";
			
			if (reader.CardPresent)
				btnConnect.Enabled = true;
			btnTransmit.Enabled = false;
			btnDisconnect.Enabled = false;

			cbProtocol.Text = "*";
		
			scard.Disconnect();
			scard = null;
		}
		
		delegate void TransmitDoneInvoker(RAPDU rapdu);
		
		void OnTransmitDone(RAPDU rapdu)
		{
			if (this.InvokeRequired)
			{
				this.BeginInvoke(new TransmitDoneInvoker(OnTransmitDone), rapdu);
				return;
			}
			
			Cursor = Cursors.Default;
			
			if (rapdu == null)
			{
				/* An error has occured */
				StatusInfo.Text = scard.LastErrorAsString;
			} else
			{
				/* OK */
				StatusInfo.Text = "";
				eRAPDU.Text = rapdu.AsString(" ");
				eSWValue.Text = String.Format("{0:X04}", rapdu.SW);
                eSWText.Text = rapdu.SWString;
			}			
			
			if (reader.CardPresent)
			{
				btnTransmit.Enabled = true;
				btnDisconnect.Enabled = true;			
			}
		}

		
		void BtnTransmitClick(object sender, EventArgs e)
		{
			Cursor = Cursors.WaitCursor;
			StatusInfo.Text = "Card communication...";
			
			eRAPDU.Text = "";
			eSWValue.Text = "";
			eSWText.Text = "";
			
			btnTransmit.Enabled = false;
			btnDisconnect.Enabled = false;			
			
			CAPDU capdu = new CAPDU(eCAPDU.Text);
			
			eCAPDU.Text = capdu.AsString(" ");
			
			scard.Transmit(capdu, new SCardChannel.TransmitDoneCallback(OnTransmitDone));
		}
		
		void MainFormFormClosed(object sender, FormClosedEventArgs e)
		{
			Cursor = Cursors.WaitCursor;
			
			if (scard != null)
			{
				scard.Disconnect();
				scard = null;
			}
			
			if (reader != null)
			{
				reader.StopMonitor();
				reader = null;
			}
			
			Cursor = Cursors.Default;
		}
		
		void CbReadersSelectedIndexChanged(object sender, EventArgs e)
		{
			if (reader != null)
			{
				reader.StopMonitor();
				reader = null;
			}

			if ((cbReaders.SelectedIndex >= 0) && (cbReaders.SelectedIndex < cbReaders.Items.Count))
			{
				reader = new SCardReader(cbReaders.Text);
				reader.StartMonitor(new SCardReader.StatusChangeCallback(OnReaderStatusChanged));
			}
		}

        private void cbProtocol_SelectedIndexChanged(object sender, EventArgs e)
        {

        }
		
	}
}
